"""entity/relation schemas to store email in an cubicweb instance

:organization: Logilab
:copyright: 2006-2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"

from cubicweb import _


# pylint: disable-msg=E0611,F0401
from yams.buildobjs import (SubjectRelation, RelationType, EntityType,
                            String, Datetime, Int, RelationDefinition)
from yams.reader import context

from cubicweb.schema import ERQLExpression


class Email(EntityType):
    """electronic mail"""
    subject = String(fulltextindexed=True)
    date = Datetime(description=_('UTC time on which the mail was sent'))
    messageid = String(required=True, indexed=True, unique=True)
    headers = String(description=_('raw headers'))

    sender = SubjectRelation('EmailAddress', cardinality='?*')
    # an email with only Bcc is acceptable, don't require any recipients
    recipients = SubjectRelation('EmailAddress')
    cc = SubjectRelation('EmailAddress')

    parts = SubjectRelation('EmailPart', cardinality='*1', composite='subject')
    attachment = SubjectRelation('File')

    reply_to = SubjectRelation('Email', cardinality='?*')
    cites = SubjectRelation('Email')
    in_thread = SubjectRelation('EmailThread', cardinality='?*')


class EmailPart(EntityType):
    """an email attachment"""
    __permissions__ = {
        'read': ('managers', 'users', 'guests',),  # XXX if E parts X, U has_read_permission E
        'add': ('managers', ERQLExpression('E parts X, U has_update_permission E'),),
        'delete': ('managers', ERQLExpression('E parts X, U has_update_permission E')),
        'update': ('managers', 'owners',),
    }

    content = String(fulltextindexed=True)
    content_format = String(required=True, maxsize=50)
    ordernum = Int(required=True)
    alternative = SubjectRelation('EmailPart', symmetric=True)


class EmailThread(EntityType):
    """discussion thread"""
    title = String(required=True, indexed=True, fulltextindexed=True)
    see_also = SubjectRelation('EmailThread')
    forked_from = SubjectRelation('EmailThread', cardinality='?*')


class parts(RelationType):
    """ """
    fulltext_container = 'subject'


class sender(RelationType):
    """ """
    inlined = True


class in_thread(RelationType):
    """ """
    inlined = True


class reply_to(RelationType):
    """ """
    inlined = True


class generated_by(RelationType):
    """mark an entity as generated from an email"""
    cardinality = '?*'
    subject = ('TrInfo',)
    object = 'Email'


# if comment is installed
if 'Comment' in context.defined:
    class comment_generated_by(RelationDefinition):
        subject = 'Comment'
        name = 'generated_by'
        object = 'Email'



"""entity classes for entity types provided by the cubicweb email package

:organization: Logilab
:copyright: 2003-2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"

import re

from logilab.common import umessage

from cubicweb.entities import AnyEntity, fetch_config, adapters
from cubicweb.predicates import is_instance

from cubicweb_email.emailcites import parse_body


class Email(AnyEntity):
    """customized class for Email entities"""
    __regid__ = 'Email'
    fetch_attrs, cw_fetch_order = fetch_config(['subject'])

    def dc_title(self):
        return self.subject

    @property
    def senderaddr(self):
        return self.sender and self.sender[0] or None

    @property
    def in_reply_to(self):
        return self.reply_to and self.reply_to[0] or None

    @property
    def thread(self):
        return self.in_thread and self.in_thread[0] or None

    def parts_in_order(self, prefered_mime_type='text/html'):
        """sort an email parts in order, selecting among alternatives according to a
        prefered mime type
        """
        parts_by_eid = {}
        alternatives = []
        result = []
        for part in self.parts:
            parts_by_eid[part.eid] = part
            if part.alternative:
                for altset in alternatives:
                    if part.eid in altset:
                        break
                else:
                    alternatives.append(set(p.eid for p in part.alternative))
                    alternatives[-1].add(part.eid)
            else:
                result.append(part)
        if alternatives:
            for altset in alternatives:
                selected = [parts_by_eid[peid] for peid in altset
                            if parts_by_eid[peid].content_format == prefered_mime_type]
                if selected:
                    selected = selected[0]
                else:
                    selected = parts_by_eid[altset.pop()]
                result.append(selected)
        return sorted(result, key=lambda x: x.ordernum)

    def references(self):
        result = set()
        message = self.umessage_headers()
        if not message:
            return result
        replyto = message.get('In-reply-to')
        if replyto:
            result.add(replyto)
        for refs in message.get_all('References', ()):
            result.update(refs.split())
        return result

    lines_rgx = re.compile(r'^Lines:\s*\d+\s*\n', re.I | re.U | re.M)
    clength_rgx = re.compile(r'^Content-Length:\s*\d+\s*\n', re.I | re.U | re.M)
    ctype_rgx = re.compile(r'^Content-Type:[^:]', re.I | re.U | re.M)

    def umessage_headers(self):
        if not self.headers:
            return None
        headers = self.lines_rgx.sub('', self.headers)
        headers = self.clength_rgx.sub('', headers)
        headers = self.ctype_rgx.sub('Content-type: text/plain; charset=utf8', headers)
        return umessage.message_from_string(headers + '\n\n')


class EmailPart(AnyEntity):
    """customized class for EmailPart entities"""
    __regid__ = 'EmailPart'

    def dc_title(self):
        return '%s (%s %s)' % (self.email.subject,
                               self._cw._('part'), self.ordernum)

    @property
    def email(self):
        return self.reverse_parts[0]

    def parent(self):
        """for breadcrumbs"""
        return self.email

    def actual_content(self):
        """return content of this part with citations and signature removed

        this method may raise `TransformError` exception if the part can't
        be displayed as text/plain.
        """
        content = self.printable_value('content', format='text/plain')
        return parse_body(content).actual_content


class EmailThread(AnyEntity):
    """customized class for EmailThread entities"""
    __regid__ = 'EmailThread'
    fetch_attrs, cw_fetch_order = fetch_config(['title'])

    def dc_title(self):
        return self.title


class EmailITreeAdapter(adapters.ITreeAdapter):
    __select__ = is_instance('Email')
    tree_relation = 'reply_to'


class EmailPartIFTIAdapter(adapters.IFTIndexableAdapter):
    """customize EmailPart IFTI adapter so we don't index pgp signature"""
    __select__ = adapters.IFTIndexableAdapter.__select__ & is_instance('EmailPart')

    def get_words(self):
        try:
            if self.entity.contenttype == 'application/pgp-signature':
                return []
        except AttributeError:
            return super(EmailPartIFTIAdapter, self).get_words()



"""automatic tests for email views"""

from cubicweb.devtools.testlib import AutomaticWebTest

from cubicweb_email.testutils import EmailValueGenerator  # noqa


class AutomaticWebTest(AutomaticWebTest):

    def to_test_etypes(self):
        return set(('Email', 'EmailPart', 'EmailThread'))

    def list_startup_views(self):
        return ()


if __name__ == '__main__':
    import unittest
    unittest.main()



# Copyright (c) 2016 LOGILAB S.A. (Paris, FRANCE).
#
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from cubicweb.devtools.fill import ValueGenerator


headers = u'''\
Return-Path: <python-projects-bounces@lists.logilab.org>
From: Benjamin Niemann <pink@odahoda.de>
To: python-projects@logilab.org
Subject: Re: [Python-projects] Pylint: Disable-msg for a block or statement?
Date: Tue, 4 Apr 2006 10:16:04 +0200
Message-Id: <200604041016.05182.pink@odahoda.de>
Sender: python-projects-bounces@lists.logilab.org

Some content'''


class EmailValueGenerator(ValueGenerator):
    """Automatic tests value generator for the Email entity

    This is required since the 'headers' attribute should be plain
    ascii only
    """

    def generate_Email_headers(self, entity, index):
        return headers



#!/usr/bin/python

from __future__ import print_function

import argparse
import mailbox
from io import BytesIO
from itertools import combinations
from email.utils import parsedate

import six

from logilab.common.umessage import message_from_file

from cwclientlib import cwproxy, cwproxy_for


class StreamMailbox(mailbox.mbox):
    """A read-only mbox format mailbox from stream."""
    _mangle_from_ = True

    def __init__(self, stream, factory=None, create=True):
        """Initialize a stream mailbox."""
        self._message_factory = mailbox.mboxMessage
        mailbox.Mailbox.__init__(self, '', factory, create)
        self._file = stream
        self._toc = None
        self._next_key = 0
        self._pending = False   # No changes require rewriting the file.
        self._locked = False
        self._file_length = None        # Used to record mailbox size


class Importer(object):
    def __init__(self, proxy, skipsign=False):
        self.proxy = proxy
        self._address_cache = {}
        self._reqs = []
        self._skipsign = skipsign

    def import_mbox_stream(self, stream):
        self._import(StreamMailbox(stream, message_from_file, create=False))

    def import_mbox(self, path):
        self._import(mailbox.mbox(path, message_from_file, create=False))

    def import_maildir(self, path):
        self._import(mailbox.Maildir(path, message_from_file, create=False))

    def _import(self, mailbox):
        for message in sorted(mailbox, key=lambda x: parsedate(x['Date'])):
            self.import_message(message)

    def import_message(self, message):
        rql = 'INSERT Email E: E headers %(headers)s'
        args = {'headers': '\n'.join('%s: %s' % header for header in message.message.items())}
        self._reqs.append((rql, args))
        emaileid = '__r%d' % (len(self._reqs) - 1)
        self._part_index = 0
        self._context = None
        self._alternatives = []
        self.import_message_parts(message, emaileid)
        try:
            resp = self.proxy.rqlio(self._reqs)
            resp.raise_for_status()
        except cwproxy.RemoteValidationError as exc:
            print("could not import message %s: %s" % (message.get('message-id'), exc.args))
        else:
            return resp.json()
        finally:
            self._reqs = []

    def import_message_parts(self, message, emaileid):
        if message.is_multipart():
            self._context = message.get_content_type().split('/')[1]
            if self._context == 'alternative':
                self._alternatives.append([])
            for part in message.get_payload():
                self.import_message_parts(part, emaileid)
            if self._context == 'alternative':
                alternatives = self._alternatives.pop()
                for eid1, eid2 in combinations(alternatives, 2):
                    self._reqs.append(('SET X alternative Y WHERE X eid %(x)s, Y eid %(y)s',
                                       {'x': eid1, 'y': eid2}))
            self._context = None
        else:
            self._import_message_part(message, emaileid)

    def _import_message_part(self, part, emaileid):
        assert not part.is_multipart()
        contenttype = part.get_content_type()
        main, sub = contenttype.split('/')
        data = part.get_payload(decode=True)
        if main == 'text':
            encoding = u'UTF-8'
        elif contenttype == 'application/pgp-signature':
            if self._skipsign:
                return
            encoding = u'ascii'
            if isinstance(data, six.binary_type):
                data = data.decode(encoding)
            # XXXX
        else:
            encoding = None
        name = part.get_filename()
        if name or main != 'text' and contenttype != 'application/pgp-signature':
            # suppose if we have a name, this is an attachement else this is a
            # part/alternative
            name = name or u'no name'
            if isinstance(data, six.text_type):
                data = data.encode('utf8')
            self._reqs.append(('INSERT File F: F data %(data)s, F data_name %(name)s, '
                               'F data_format %(ctype)s, F data_encoding %(encoding)s',
                               {'data': BytesIO(data),
                                'name': name,
                                'ctype': contenttype,
                                'encoding': encoding}))
            parteid = '__r%d' % (len(self._reqs) - 1)
            self._reqs.append(('SET X attachment Y WHERE X eid %(x)s, Y eid %(y)s',
                               {'x': emaileid, 'y': parteid}))
        else:
            self._part_index += 1
            self._reqs.append(('INSERT EmailPart P: P content %(data)s, P content_format %(ctype)s, P ordernum %(idx)s',
                               {'data': data, 'ctype': contenttype, 'idx': self._part_index}))
            parteid = '__r%d' % (len(self._reqs) - 1)
            self._reqs.append(('SET X parts Y WHERE X eid %(x)s, Y eid %(y)s',
                               {'x': emaileid, 'y': parteid}))
            if self._context == 'alternative':
                self._alternatives[-1].append(parteid)


def main():
    parser = argparse.ArgumentParser(
        description='import email into a cubicweb instance')
    parser.add_argument(
        'endpoint', metavar='endpoint', type=str,
        help='cwclientlib endpoint of the cubicweb instance')
    parser.add_argument(
        'mbox', metavar='mailbox', type=argparse.FileType('rb'),
        nargs='+', help='mbox files to import (- for stdin)')
    parser.add_argument(
        '-S', '--skipsign', action='store_true',
        default=False,
        help='do not create EmailPart for PGP/GPG signatures')
    args = parser.parse_args()

    proxy = cwproxy_for(args.endpoint)

    importer = Importer(proxy, skipsign=args.skipsign)
    for f in args.mbox:
        try:
            f.seek(0)
        except IOError:
            buf = BytesIO()
            buf.writelines(f)
            buf.seek(0)
            f = buf
        importer.import_mbox_stream(f)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
import unittest

from cubicweb_email.emailcites import parse_body


class ParseBodyTC(unittest.TestCase):

    def test_remove_sig(self):
        parsedmsg = parse_body('''
Des gens qui se sont mis sur le secteur erp avec du zope, du php, de postgresql et du debian : pimentech.

jpl/crm/compta/paie intégré : http://www.pimentech.fr/solutions/pimenstud
gestion/compta :              http://www.pimentech.fr/solutions/pimengest
travail collaboratif :        http://www.pimentech.fr/solutions/notesgroup
CRM :                         http://www.pimentech.fr/solutions/crm

--
Alexandre Fayolle
''')
        self.assertMultiLineEqual(parsedmsg.actual_content, '''Des gens qui se sont mis sur le secteur erp avec du zope, du php, de postgresql et du debian : pimentech.

jpl/crm/compta/paie intégré : http://www.pimentech.fr/solutions/pimenstud
gestion/compta :              http://www.pimentech.fr/solutions/pimengest
travail collaboratif :        http://www.pimentech.fr/solutions/notesgroup
CRM :                         http://www.pimentech.fr/solutions/crm''')
        self.assertMultiLineEqual(parsedmsg.cited_content, '')

    def test_remove_cites(self):
        parsedmsg = parse_body('''
On Tue, Jan 08, 2008 at 10:08:46AM +0100, Alexandre Fayolle wrote:
> Des gens qui se sont mis sur le secteur erp avec du zope, du php, de postgresql et du debian : pimentech.
>
> jpl/crm/compta/paie intégré : http://www.pimentech.fr/solutions/pimenstud
> gestion/compta :              http://www.pimentech.fr/solutions/pimengest
> travail collaboratif :        http://www.pimentech.fr/solutions/notesgroup
> CRM :                         http://www.pimentech.fr/solutions/crmi

leur solution CRM est même basé sur Zope...

--
Sylvain Thénault                               LOGILAB, Paris (France)
Formations Python, Zope, Plone, Debian:  http://www.logilab.fr/formations
Développement logiciel sur mesure:       http://www.logilab.fr/services
Python et calcul scientifique:           http://www.logilab.fr/science
''')
        self.assertMultiLineEqual(parsedmsg.actual_content, '''leur solution CRM est même basé sur Zope...''')
        self.assertMultiLineEqual(parsedmsg.cited_content, '''On Tue, Jan 08, 2008 at 10:08:46AM +0100, Alexandre Fayolle wrote:
Des gens qui se sont mis sur le secteur erp avec du zope, du php, de postgresql et du debian : pimentech.
jpl/crm/compta/paie intégré : http://www.pimentech.fr/solutions/pimenstud
gestion/compta :              http://www.pimentech.fr/solutions/pimengest
travail collaboratif :        http://www.pimentech.fr/solutions/notesgroup
CRM :                         http://www.pimentech.fr/solutions/crmi''')

    def test_ml(self):
        parsedmsg = parse_body('''
On Wed, Jan 23, 2008 at 06:48:22PM +0100, Maarten ter Huurne wrote:
> I wrote on 2008-01-23 06:27:07 PM:
>
> [Lotus Notes ate my whitespace, here is the proper code layout, I hope]
>
> > ===
> > class SomeClass(object):
> >
> >     def __m(self):
> >         pass
> >
> >     SomeClass().__m()
> > ===
>

Hi Marteen,

did you actually execute that code ?

> Bye,
> Maarten

> _______________________________________________
> Python-Projects mailing list
> Python-Projects@lists.logilab.org
> http://lists.logilab.org/mailman/listinfo/python-projects
_______________________________________________
Python-Projects mailing list
Python-Projects@lists.logilab.org
http://lists.logilab.org/mailman/listinfo/python-projects
''')
        self.assertMultiLineEqual(parsedmsg.actual_content, '''Hi Marteen,

did you actually execute that code ?
_______________________________________________
Python-Projects mailing list
Python-Projects@lists.logilab.org
http://lists.logilab.org/mailman/listinfo/python-projects''')
        self.assertMultiLineEqual(parsedmsg.cited_content, '''On Wed, Jan 23, 2008 at 06:48:22PM +0100, Maarten ter Huurne wrote:
I wrote on 2008-01-23 06:27:07 PM:
[Lotus Notes ate my whitespace, here is the proper code layout, I hope]
===
class SomeClass(object):
def __m(self):
pass
SomeClass().__m()
===
Bye,
Maarten
_______________________________________________
Python-Projects mailing list
Python-Projects@lists.logilab.org
http://lists.logilab.org/mailman/listinfo/python-projects''')


if __name__ == '__main__':
    unittest.main()



# -*- coding: iso-8859-1 -*-
import sys
from os.path import join, dirname, abspath
import unittest

from socket import gethostname
from io import BytesIO

from logilab.common.testlib import unittest_main
from cwclientlib import cwproxy

from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.pyramid.test import PyramidCWTest
from cubicweb.mail import construct_message_id

from cwemail.mboximport import Importer


MSG = u'''From sthenault@free.fr  Tue Jan 23 15:21:10 2007
Return-Path: <sthenault@free.fr>
X-Original-To: Sylvain.Thenault@logilab.fr
Delivered-To: syt@logilab.fr
Received: from tucana.logilab.fr (tucana.logilab.fr [172.17.0.4])
        by orion.logilab.fr (Postfix) with ESMTP id E4D79EAA37
        for <Sylvain.Thenault@logilab.fr>; Tue, 23 Jan 2007 15:21:09 +0100 (CET)
Received: from smtp2-g19.free.fr (smtp2-g19.free.fr [212.27.42.28])
        by tucana.logilab.fr (Postfix) with ESMTP id 23B3D7140CE
        for <Sylvain.Thenault@logilab.fr>; Tue, 23 Jan 2007 15:21:08 +0100 (CET)
Received: from nor75-16-82-239-114-67.fbx.proxad.net
        (cap31-2-82-224-153-77.fbx.proxad.net [82.224.153.77])
        by smtp2-g19.free.fr (Postfix) with ESMTP id B057D7D54
        for <Sylvain.Thenault@logilab.fr>; Tue, 23 Jan 2007 15:21:07 +0100 (CET)
MIME-Version: 1.0
Content-Type: text/plain; charset="utf-8"
Content-Transfer-Encoding: base64
Subject: =?utf-8?q?new_comment_for_bug_avoir_un_champ_=22python_compat=22_sur_les_?=
        =?utf-8?q?projets_=3F?=
From: =?utf-8?q?Sylvain_Th=C3=A9nault?= <Sylvain.Thenault@logilab.fr>
Reply-To: =?utf-8?q?Sylvain_Th=C3=A9nault?= <Sylvain.Thenault@logilab.fr>
X-ERUDI: jpl
To: Sylvain.Thenault@logilab.fr
Message-id: <iblaoliejr>
References: <rooteid=15058&eid=17244&timestamp=1169562059@%s.%s>
Date: Tue, 23 Jan 2007 15:21:07 +0100 (CET)
X-Spambayes-Classification: ham; 0.00
Status: RO
Content-Length: 122
Lines: 2
'''


class ChangeStateHooksTC(CubicWebTC):

    def setup_database(self):
        with self.admin_access.client_cnx() as cnx:
            cnx.create_entity('BlogEntry', title=u"une news !", content=u"cubicweb c'est beau")
            cnx.create_entity('EmailAddress', address=u'devel@logilab.fr', alias=u'devel')
            cnx.commit()
        self.msg = MSG % (self.vreg.config.appid, gethostname())

    def test_email_change_state(self):
        with self.admin_access.client_cnx() as cnx:
            u = cnx.execute('INSERT CWUser X: X login "toto", X upassword "sosafe", X in_group G '
                            'WHERE G name "users"').get_entity(0, 0)
            subject = u':deactivate: %s' % u.eid
            cnx.execute('INSERT Email X: '
                        'X headers %(headers)s, X subject %(subject)s,'
                        'X messageid "hop", X sender EA',
                        {'subject': subject, 'headers': self.msg})[0][0]
            cnx.commit()
            userstate = cnx.execute('Any SN WHERE X in_state S, S name SN, X eid %(x)s',
                                    {'x': u.eid})[0][0]
            self.assertEqual(userstate, 'deactivated')
            self.assertEqual(u.cw_adapt_to('IWorkflowable').latest_trinfo().creator.login, 'admin')

    def test_email_change_state_trinfo_owner(self):
        with self.admin_access.client_cnx() as cnx:
            cnx.execute('SET U use_email E WHERE U login "anon"')
            u = cnx.execute('INSERT CWUser X: X login "toto", X upassword "sosafe", X in_group G '
                            'WHERE G name "users"').get_entity(0, 0)
            subject = u':deactivate: %s' % u.eid
            cnx.execute('INSERT Email X: '
                        'X headers %(headers)s, X subject %(subject)s,'
                        'X messageid "hop", X sender EA',
                        {'subject': subject, 'headers': self.msg})[0][0]
            cnx.commit()
            userstate = cnx.execute('Any SN WHERE X in_state S, S name SN, X eid %(x)s',
                                    {'x': u.eid})[0][0]
            self.assertEqual(userstate, 'deactivated')
            self.assertEqual(u.cw_adapt_to('IWorkflowable').latest_trinfo().creator.login, 'anon')

    def test_email_change_state_wrong_eid(self):
        with self.admin_access.client_cnx() as cnx:
            ueid = cnx.execute('INSERT CWUser X: X login "toto", X upassword "sosafe", X in_group G '
                               'WHERE G name "users"')[0][0]
            subject = u':deactivate: %s' % 10
            cnx.execute('INSERT Email X: '
                        'X headers %(headers)s, X subject %(subject)s,'
                        'X messageid "hop", X sender EA',
                        {'subject': subject, 'headers': self.msg})[0][0]
            cnx.commit()
            userstate = cnx.execute('Any SN WHERE X in_state S, S name SN, X eid %(x)s',
                                    {'x': ueid})[0][0]
            self.assertEqual(userstate, 'activated')

    def test_email_change_state_unexistant_eid(self):
        with self.admin_access.client_cnx() as cnx:
            ueid = cnx.execute('INSERT CWUser X: X login "toto", X upassword "sosafe", X in_group G '
                               'WHERE G name "users"')[0][0]
            subject = u':deactivate: %s' % 999999
            cnx.execute('INSERT Email X: '
                        'X headers %(headers)s, X subject %(subject)s,'
                        'X messageid "hop", X sender EA',
                        {'subject': subject, 'headers': self.msg})[0][0]
            cnx.commit()
            userstate = cnx.execute('Any SN WHERE X in_state S, S name SN, X eid %(x)s',
                                    {'x': ueid})[0][0]
            self.assertEqual(userstate, 'activated')

    def test_email_change_state_unexistant_transition(self):
        with self.admin_access.client_cnx() as cnx:
            ueid = cnx.execute('INSERT CWUser X: X login "toto", X upassword "sosafe", X in_group G '
                               'WHERE G name "users"')[0][0]
            subject = u':blabla: %s' % ueid
            cnx.execute('INSERT Email X: '
                        'X headers %(headers)s, X subject %(subject)s,'
                        'X messageid "hop", X sender EA',
                        {'subject': subject, 'headers': self.msg})[0][0]
            cnx.commit()
            userstate = cnx.execute('Any SN WHERE X in_state S, S name SN, X eid %(x)s',
                                    {'x': ueid})[0][0]
            self.assertEqual(userstate, 'activated')


MSG2 = u'''MIME-Version: 1.0
Content-Type: text/plain; charset="utf-8"
Content-Transfer-Encoding: base64
Subject: =?utf-8?q?new_comment_for_bug_avoir_un_champ_=22python_compat=22_sur_les_?=
        =?utf-8?q?projets_=3F?=
From: =?utf-8?q?Sylvain_Th=C3=A9nault?= <Sylvain.Thenault@logilab.fr>
Reply-To: =?utf-8?q?Sylvain_Th=C3=A9nault?= <Sylvain.Thenault@logilab.fr>
X-ERUDI: jpl
To: Sylvain.Thenault@logilab.fr
Message-id: <iblaoliejr>
References: %s
Date: Tue, 23 Jan 2007 15:21:07 +0100 (CET)
X-Spambayes-Classification: ham; 0.00
Status: RO
Content-Length: 122
Lines: 2

hop
hop
'''


class ReplyCommentHooksTC(PyramidCWTest):
    token_id = u'mboximport'

    def setup_database(self):
        with self.admin_access.client_cnx() as cnx:
            self.b = cnx.create_entity('BlogEntry', title=u"une news !",
                                       content=u"cubicweb c'est beau").eid
            e = cnx.create_entity('EmailAddress',
                                  address=u'sylvain.thenault@logilab.fr',
                                  alias=u'syt')
            cnx.execute('SET X use_email E WHERE X login "anon", E eid %(e)s',
                        {'e': e.eid})
            cnx.create_entity('AuthToken', enabled=True, id=self.token_id,
                              token_for_user=cnx.user)
            cnx.commit()
            self.secret = cnx.execute('String X WHERE T token X, T id %(id)s',
                                      {'id': self.token_id})[0][0]

    @unittest.skip('XXX this should be migrated to proper PyramidCWTest')
    def test_comment_created(self):
        msg = open(join(self.datadir, 'reply.mbox'), 'rb').read() % \
            construct_message_id(self.vreg.config.appid, self.b, False)
        mi = Importer(cwproxy.CWProxy(self.config['base-url'],
                                      auth=cwproxy.SignedRequestAuth(self.token_id,
                                                                     self.secret)))
        mi.import_mbox_stream(BytesIO(msg.encode('utf-8')))
        with self.admin_access.client_cnx() as cnx:
            self.assertTrue(cnx.execute('Email X'))
            b = cnx.entity_from_eid(self.b)
            self.assertTrue(b.reverse_comments)
            c = b.reverse_comments[0]
            self.assertEqual(c.content_format, 'text/plain')
            self.assertEqual(c.content, u'''New ticket for project erudi-email :

#22742 - erudi-email #22742: commentaires de mailbot
====================

description
-----------
Sur http://intranet.logilab.fr/jpl/ticket/22309 on voit un commentaire sign\xe9
mailbot. Ce serait mieux d'utiliser l'auteur du mail.

submitter
---------
alf

URL
---
http://intranet.logilab.fr/jpl/ticket/22742
(project URL: http://intranet.logilab.fr/jpl/project/erudi-email)''')
            self.assertEqual(c.creator.login, 'anon')
            self.assertEqual(sorted(u.login for u in c.owned_by),
                             ['admin', 'anon'])


if __name__ == '__main__':
    unittest_main()



from cubicweb.predicates import is_instance
from cubicweb.web.views import ibreadcrumbs, uicfg

_pvs = uicfg.primaryview_section
_pvs.tag_object_of(('Email', 'in_thread', 'EmailThread'), 'hidden')

# XXX fix cw.web.views.emailaddress.EmailAddressPrimaryView view
# using proper view and uicfg config
_pvs.tag_object_of(('*', 'sender', 'EmailAddress'), 'hidden')
_pvs.tag_object_of(('*', 'recipients', 'EmailAddress'), 'hidden')
_pvs.tag_object_of(('*', 'cc', 'EmailAddress'), 'hidden')

_afs = uicfg.autoform_section
_afs.tag_object_of(('*', 'sender', 'EmailAddress'), 'main', 'hidden')
_afs.tag_object_of(('*', 'recipients', 'EmailAddress'), 'main', 'hidden')
_afs.tag_object_of(('*', 'cc', 'EmailAddress'), 'main', 'hidden')
_afs.tag_subject_of(('*', 'generated_by', 'Email'), 'main', 'hidden')


class EmailPartIBreadCrumbsAdapter(ibreadcrumbs.IBreadCrumbsAdapter):
    __select__ = is_instance('EmailPart')

    def parent_entity(self):
        return self.entity.email



# pylint: disable=W0622
"""cubicweb-email packaging information"""

modname = 'email'
distname = "cubicweb-%s" % modname

numversion = (1, 13, 1)
version = '.'.join(str(num) for num in numversion)

license = 'LGPL'
author = "Logilab"
author_email = "contact@logilab.fr"
web = 'http://www.cubicweb.org/project/%s' % distname
description = "email component for the CubicWeb framework"
classifiers = [
    'Environment :: Web Environment',
    'Framework :: CubicWeb',
    'Programming Language :: Python',
    'Programming Language :: JavaScript',
]
__depends__ = {
    'cubicweb': '>= 3.24',
    'cubicweb-file': '>= 1.9.0',
    'cwclientlib': '>= 0.3.1',
}



"""Specific views for email related entities

:organization: Logilab
:copyright: 2003-2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"

from cubicweb import _

from logilab.mtconverter import xml_escape

from cubicweb.predicates import is_instance
from cubicweb.uilib import soup2xhtml
from cubicweb.view import EntityView
from cubicweb.web import formwidgets
from cubicweb.web.views import baseviews, primary, uicfg

for rtype in ('sender', 'recipients', 'cc', 'parts'):
    uicfg.primaryview_section.tag_subject_of(('Email', rtype, '*'), 'hidden')

uicfg.autoform_field_kwargs.tag_attribute(('Email', 'subject'),
                                          {'widget': formwidgets.TextInput})

uicfg.actionbox_appearsin_addmenu.tag_subject_of(('Email', 'attachment', '*'),
                                                 True)

uicfg.actionbox_appearsin_addmenu.tag_object_of(('EmailThread', 'forked_from', 'EmailThread'),
                                                True)


def formated_sender(email):
    if email.sender:
        return email.sender[0].view('oneline')
    # sender address has been removed, look in email's headers
    message = email.umessage_headers()
    if message:
        return xml_escape(message.get('From', ''))
    return email._cw._('unknown sender')


class EmailPrimaryView(primary.PrimaryView):
    __select__ = is_instance('Email')

    def render_entity_attributes(self, entity):
        self.w(u'<div class="emailheader"><table>')
        self.w(u'<tr><td>%s</td><td>%s</td></tr>' %
               (self._cw._('From'), formated_sender(entity)))
        self.w(u'<tr><td>%s</td><td>%s</td></tr>' %
               (self._cw._('To'), ', '.join(ea.view('oneline') for ea in entity.recipients)))
        if entity.cc:
            self.w(u'<tr><td>%s</td><td>%s</td></tr>' %
                   (self._cw._('CC'), ', '.join(ea.view('oneline') for ea in entity.cc)))
        self.w(u'<tr><td>%s</td><td>%s</td></tr>' %
               (self._cw._('Date'), self._cw.format_date(entity.date, time=True)))
        self.w(u'<tr><td>%s</td><td>%s</td></tr>' %
               (self._cw._('Subject'), xml_escape(entity.subject)))
        self.w(u'</table></div><div class="emailcontent">')
        for part in entity.parts_in_order():
            content, mime = part.content, part.content_format
            if mime == 'text/html':
                content = soup2xhtml(content, self._cw.encoding)
            elif 'pgp-signature' in mime:
                content = entity._cw_mtc_transform(content, mime, 'text/html', self._cw.encoding)
            elif mime != 'text/xhtml':
                content = xml_escape(content)
                if mime == 'text/plain':
                    content = content.replace('\n', '<br/>').replace(' ', '&nbsp;')
            # XXX some headers to skip if html ?
            self.w(content)
            self.w(u'<br class="partseparator"/>')
        self.w(u'</div>')

    def render_entity_title(self, entity):
        self.w(u'<h1><span class="etype">%s</span> %s</h1>'
               % (entity.dc_type().capitalize(), xml_escape(entity.dc_title())))


class EmailHeadersView(baseviews.EntityView):
    """display email's headers"""
    __regid__ = 'headers'
    __select__ = is_instance('Email')
    title = _('headers')
    templatable = False
    content_type = 'text/plain'

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        self.w(entity.headers)


class EmailOneLineView(baseviews.OneLineView):
    """short view usable in the context of the email sender/recipient (in which
    case the caller should specify its context eid) or outside any context
    """
    __select__ = is_instance('Email')
    title = _('oneline')

    def cell_call(self, row, col, contexteid=None):
        entity = self.cw_rset.get_entity(row, col)
        self.w(u'<div class="email">')
        self.w(u'<i>%s&nbsp;%s</i> ' % (
            self._cw._('email_date'), self._cw.format_date(entity.date, time=True)))
        sender = entity.senderaddr
        if sender is None or contexteid != sender.eid:
            self.w(u'<b>%s</b>&nbsp;%s '
                   % (self._cw._('email_from'), formated_sender(entity)))
        if contexteid not in (r.eid for r in entity.recipients):
            recipients = ', '.join(r.view('oneline') for r in entity.recipients)
            self.w(u'<b>%s</b>&nbsp;%s'
                   % (self._cw._('email_to'), recipients))
        self.w(u'<br/>\n<a href="%s">%s</a>' % (
            xml_escape(entity.absolute_url()), xml_escape(entity.subject)))
        self.w(u'</div>')


class EmailOutOfContextView(EmailOneLineView):
    """short view outside the context of the email"""
    __regid__ = 'outofcontext'
    title = _('out of context')


class EmailInContextView(EmailOneLineView):
    """short view inside the context of the email"""
    __regid__ = 'incontext'


class EmailTreeItemView(EmailOutOfContextView):
    __regid__ = 'treeitem'
    title = None


class EmailPartOutOfContextView(baseviews.OutOfContextView):
    """out of context an email part is redirecting to related email view"""
    __select__ = is_instance('EmailPart')

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        entity.reverse_parts[0].view('outofcontext', w=self.w)


class EmailThreadView(EntityView):
    __regid__ = 'emailthread'
    __select__ = is_instance('EmailThread')

    def entity_call(self, entity):
        # get top level emails in this thread (ie message which are not a reply
        # of a message in this thread)
        #
        # Warn: adding Y in_thread E changes the meaning of the query since it joins
        # with messages which are not a direct reply (eg if A reply_to B, B reply_to C
        # B is also retreived since it's not a reply of C
        #
        # XXX  union with
        #   DISTINCT Any X,D ORDERBY D WHERE X date D, X in_thread E, X reply_to Y,
        #   NOT Y in_thread E, E eid %(x)s'
        # to get message which are a reply of a message in another thread ?
        # we may get duplicates in this case
        rset = self._cw.execute('DISTINCT Any X,D ORDERBY D '
                                'WHERE X date D, X in_thread E, '
                                'NOT X reply_to Y, E eid %(x)s',
                                {'x': entity.eid})
        if rset:
            self.w(u'<ul>')
            self.w(u'\n'.join(email.view('tree') for email in rset.entities()))
            self.w(u'</ul>')


class EmailThreadPrimaryView(primary.PrimaryView):
    __select__ = is_instance('EmailThread')

    def cell_call(self, row, col):
        entity = self.cw_rset.complete_entity(row, col)
        self.w(u'<h1>%s</h1>' % xml_escape(entity.title))
        entity.view('emailthread', w=self.w)



"""unit tests for email mbox import functionnality"""

import unittest
from os import mkdir
from os.path import join, exists
from io import BytesIO

from cwclientlib.cwproxy import CWProxy, SignedRequestAuth

from cwemail.mboximport import Importer


from cubicweb.pyramid.test import PyramidCWTest


class MBOXImporterTC(PyramidCWTest):
    token_id = u'mboximport'
    test_db_id = u'mboximport'

    @classmethod
    def pre_setup_database(cls, cnx, config):
        cnx.create_entity('AuthToken', id=cls.token_id,
                          enabled=True,
                          token_for_user=cnx.user.eid)
        cnx.commit()

    def setUp(self):
        super(MBOXImporterTC, self).setUp()
        with self.admin_access.client_cnx() as cnx:
            token = cnx.find('AuthToken', id=self.token_id).one()
            self.admineid = cnx.user.eid
            self.token = token.token
        auth = SignedRequestAuth(token_id=self.token_id,
                                 secret=self.token)
        self.client = CWProxy(base_url=self.config['base-url'],
                              auth=auth)

    @unittest.skip('should be migrated to proper PyramidCWTest')
    def test_mbox_format(self):
        mi = Importer(self.client)
        mi.import_mbox(self.datapath('mbox'))
        with self.admin_access.client_cnx() as cnx:
            self._test_base(cnx)

    @unittest.skip('should be migrated to proper PyramidCWTest')
    def test_maildir_format(self):
        mi = Importer(self.client)
        maildirpath = self.datapath('maildir')
        newpath = join(maildirpath, 'new')
        if not exists(newpath):
            mkdir(newpath)
        mi.import_maildir(maildirpath)
        with self.admin_access.client_cnx() as cnx:
            self._test_base(cnx)

    def _test_base(self, cnx):
        rset = cnx.execute('Any X, S ORDERBY S WHERE X is Email, X subject S')
        self.assertEqual(len(rset), 2, rset.rows)
        email = rset.get_entity(0, 0)
        self.assertEqual(email.subject, 'Re: [Python-projects] Pylint: Disable-msg for a block or statement?')
        self.assertEqual(email.sender[0].address, 'pink@odahoda.de')
        self.assertEqual(email.sender[0].alias, 'Benjamin Niemann')
        self.assertEqual([r.address for r in email.recipients], ['python-projects@logilab.org'])
        self.assertEqual([r.address for r in email.cc], [])
        self.assertEqual(email.in_thread[0].title, '[Python-projects] Pylint: Disable-msg for a block or statement?')
        self.assertEqual(sorted([(f.data_name, f.data_format) for f in email.attachment]),
                         [(u'astng.patch', u'text/x-diff'), (u'pylint.patch', u'text/x-diff')])
        assert not self.vreg.schema['parts'].inlined
        self.assertEqual(len(cnx.execute('Any P WHERE E parts P, E eid %s' % email.eid)), 2)
        self.assertEqual(len(email.parts), 2)
        part1, part2 = email.parts_in_order()
        self.assertEqual(part1.content_format, 'text/plain')
        self.assertEqual(part2.content_format, 'text/plain')
        self.assertEqual(part2.content, '''_______________________________________________
Python-Projects mailing list
Python-Projects@lists.logilab.org
http://lists.logilab.org/mailman/listinfo/python-projects''')

        email = rset.get_entity(1, 0)
        self.assertEqual(email.subject, '[Python-projects] pylint: False positive about field initialisation')
        self.assertEqual(email.sender[0].address, 'maarten.ter.huurne@philips.com')
        self.assertEqual(email.sender[0].alias, 'Maarten ter Huurne')
        self.assertEqual([r.address for r in email.recipients], ['python-projects@lists.logilab.org'])
        self.assertEqual([r.address for r in email.cc], [])
        self.assertEqual(email.in_thread[0].title,
                         '[Python-projects] pylint: False positive about field initialisation')
        self.assertEqual([(f.data_name, f.data_format) for f in email.attachment], [])
        self.assertEqual(len(email.parts), 3)
        self.assertEqual(sorted([r.content_format for r in email.parts]), [u'text/html', u'text/plain', u'text/plain'])
        part1, part2 = email.parts_in_order()
        self.assertEqual(part1.content_format, 'text/html')
        self.assertEqual(part2.content_format, 'text/plain')
        self.assertEqual(part2.content, '''_______________________________________________
Python-Projects mailing list
Python-Projects@lists.logilab.org
http://lists.logilab.org/mailman/listinfo/python-projects''')
        part1, part2 = email.parts_in_order('text/plain')
        self.assertEqual(part1.content_format, 'text/plain')
        self.assertEqual(part2.content_format, 'text/plain')

    NOSUBJECT = b"""
From python-projects-bounces@lists.logilab.org  Tue Mar 28 17:06:41 2006
Return-Path: <python-projects-bounces@lists.logilab.org>
X-Original-To: sylvain.thenault@logilab.fr
Delivered-To: syt@logilab.fr
To: python-projects@lists.logilab.org
MIME-Version: 1.0
X-Mailer: Lotus Notes Release 6.0.3 September 26, 2003
Message-ID: <OFE9ED18DA.E3F2A23A-ONC125713F.00521FE5-C125713F.0052FE93@philips.com>
References: <xxx@blabla>
From: Maarten ter Huurne <MAARTEN.TER.HUURNE@PHILIPS.COM>
Date: Tue, 28 Mar 2006 17:05:23 +0200
X-MIMETrack: Serialize by Router on ehvrmh02/H/SERVER/PHILIPS(Release
\t6.5.3FP1HF291 | September 19, 2005) at 28/03/2006 17:05:23,
\tSerialize complete at 28/03/2006 17:05:23
X-BeenThere: python-projects@lists.logilab.org
X-Mailman-Version: 2.1.5
Precedence: list
Content-Type: multipart/mixed; boundary="===============1708445001=="
Mime-version: 1.0
Sender: python-projects-bounces@lists.logilab.org
Errors-To: python-projects-bounces@lists.logilab.org
X-Spambayes-Classification: ham; 0.00
Status: RO
Content-Length: 2915
Lines: 82

This is a multipart message in MIME format.
--===============1708445001==
Content-Type: multipart/alternative;
\tboundary="=_alternative 0052FE92C125713F_="

This is a multipart message in MIME format.
--=_alternative 0052FE92C125713F_=
Content-Type: text/plain; charset="US-ASCII"

Hi,

On the following program:
===
class C:
    def __init__(self):
        self.set(self, 'abc')
    def set(self, value):
        self.__value = value
        self.__length = len(value)
===
pylint reports:
===
W0201:  5:C.set: Attribute '__value' defined outside __init__
W0201:  6:C.set: Attribute '__length' defined outside __init__
===

Although strictly speaking they are indeed defined outside __init__, these
fields are guaranteed to be initialised when an object of type C is
constructed. It would be useful if pylint could recognise situations like
this one and not issue this warning.

Bye,
                Maarten

--=_alternative 0052FE92C125713F_=
Content-Type: text/html; charset="US-ASCII"


<br><font size=2 face="sans-serif">Hi,</font>
<br>
<br><font size=2 face="sans-serif">On the following program:</font>
<br><font size=2 face="sans-serif">===</font>
<br><font size=2 face="sans-serif">class C:</font>
<br><font size=2 face="sans-serif">&nbsp; &nbsp; def __init__(self):</font>
<br><font size=2 face="sans-serif">&nbsp; &nbsp; &nbsp; &nbsp; self.set(self,
'abc')</font>
<br><font size=2 face="sans-serif">&nbsp; &nbsp; def set(self, value):</font>
<br><font size=2 face="sans-serif">&nbsp; &nbsp; &nbsp; &nbsp; self.__value
= value</font>
<br><font size=2 face="sans-serif">&nbsp; &nbsp; &nbsp; &nbsp; self.__length
= len(value)</font>
<br><font size=2 face="sans-serif">===</font>
<br><font size=2 face="sans-serif">pylint reports:</font>
<br><font size=2 face="sans-serif">===</font>
<br><font size=2 face="sans-serif">W0201: &nbsp;5:C.set: Attribute '__value'
defined outside __init__</font>
<br><font size=2 face="sans-serif">W0201: &nbsp;6:C.set: Attribute '__length'
defined outside __init__</font>
<br><font size=2 face="sans-serif">===</font>
<br>
<br><font size=2 face="sans-serif">Although strictly speaking they are
indeed defined outside __init__, these fields are guaranteed to be initialised
when an object of type C is constructed. It would be useful if pylint could
recognise situations like this one and not issue this warning.</font>
<br>
<br><font size=2 face="sans-serif">Bye,</font>
<br><font size=2 face="sans-serif">&nbsp; &nbsp; &nbsp; &nbsp; &nbsp;
&nbsp; &nbsp; &nbsp; Maarten</font>
<br>
--=_alternative 0052FE92C125713F_=--

--===============1708445001==
Content-Type: text/plain; charset="us-ascii"
MIME-Version: 1.0
Content-Transfer-Encoding: 7bit
Content-Disposition: inline

_______________________________________________
Python-Projects mailing list
Python-Projects@lists.logilab.org
http://lists.logilab.org/mailman/listinfo/python-projects
--===============1708445001==--
"""

    @unittest.skip('should be migrated to proper PyramidCWTest')
    def test_no_subject(self):
        mi = Importer(self.client)
        mi.import_mbox_stream(BytesIO(self.NOSUBJECT))
        with self.admin_access.client_cnx() as cnx:
            rset = cnx.execute('Any X ORDERBY S WHERE X is Email, X subject S')
            self.assertEqual(len(rset), 1)
            email = rset.get_entity(0, 0)
            self.assertEqual(email.subject, '(no subject)')
            self.assertEqual(email.references(), set(('<xxx@blabla>',)))

    @unittest.skip('should be migrated to proper PyramidCWTest')
    def test_double_import(self):
        mi = Importer(self.client)
        mi.import_mbox_stream(BytesIO(self.NOSUBJECT))
        mi.import_mbox_stream(BytesIO(self.NOSUBJECT))
        with self.admin_access.client_cnx() as cnx:
            rset = cnx.execute('Any X, S ORDERBY S WHERE X is Email, X subject S')
            self.assertEqual(len(rset), 1, rset.rows)

    @unittest.skip('should be migrated to proper PyramidCWTest')
    def test_address_detection(self):
        with self.admin_access.client_cnx() as cnx:
            # test both case normalization and canonicalization of email address
            eid1 = cnx.execute('INSERT EmailAddress X: X address "maarten.ter.huurne@philips.com"')[0][0]
            eid2 = cnx.execute('INSERT EmailAddress X: X address "maarten@philips.com"')[0][0]
            cnx.execute('SET X prefered_form Y WHERE X eid %s, Y eid %s' % (eid1, eid2))
            cnx.commit()
        mi = Importer(self.client)
        mi.import_mbox_stream(BytesIO(self.NOSUBJECT))
        with self.admin_access.client_cnx() as cnx:
            rset = cnx.execute('Any X ORDERBY S WHERE X is Email, X subject S')
            self.assertEqual(len(rset), 1)
            email = rset.get_entity(0, 0)
            self.assertEqual(email.sender[0].address, 'maarten@philips.com')


if __name__ == '__main__':
    unittest.main()



"""some utilities to process email information

:organization: Logilab
:copyright: 2007-2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"


def parse_body(body, quote='>'):
    """return a `ParsedMessage` instance where actual content is separated from
    cited content
    """
    res = []
    level = 0
    current = []
    seenlevels = set()
    for line in body.splitlines(True):
        newlevel = 0
        while line.startswith(quote):
            newlevel += 1
            line = line[1:].lstrip()
        if line.strip() == '--':
            break  # ignore following signature
        if newlevel != level:
            # detect lines such as
            #   On Wed, Jan 23, 2008 at 06:48:22PM +0100, machin wrote:
            # preceding citation
            if (
                newlevel not in seenlevels and res and res[-1][1]
                and res[-1][1][-1].rstrip().endswith(':')
            ):
                current.insert(0, res[-1][1].pop())
            seenlevels.add(newlevel)
            res.append((level, current))
            current = []
        level = newlevel
        current.append(line)
    res.append((level, current))
    pmsg = ParsedMessage()
    for level, lines in res:
        para = ''.join(lines).strip()
        if not para:
            continue
        if level > 0:
            pmsg.cites.append(para)
        else:
            pmsg.content.append(para)
    return pmsg


class ParsedMessage(object):
    def __init__(self):
        self.content = []
        self.cites = []

    @property
    def actual_content(self):
        return '\n'.join(self.content)

    @property
    def cited_content(self):
        return '\n'.join(self.cites)

# import sys
# from mailbox import UnixMailbox
# from maboa.umessage import message_from_file
# from cubicweb.web.uilib import remove_html_tags

# def ignore_cite(para):
#     para = para.lower()
#     if para == u'ok':
#         return True
#     return False

# class CitesDetector:
#     """XXX TODO
#     * unitttest :)
#     * limit search citations according to UTC date
#     * skip block under a reasonable size (['ok'])
#     * normalize block text
#     * enhance block detection (extract quote string in a first phase)
#     """
#     def _get_body(self, message):
#         body = []
#         for part in message.walk():
#             if part.get_content_maintype() == 'text':
#                 payload = part.get_payload(decode=True)
#                 if part.get_content_type().endswith('html'):
#                     payload = remove_html_tags(payload)
#                 body.append(payload)
#             elif part.get_content_type() == 'application/octet-stream':
#                 body.append(u'attached file: %s (%s)' %
#                             (part.get_filename(''), part.get_content_type()))
#             elif part.get_content_type() == 'multipart/mixed':
#                 pass
#         return u'\n\n--- part ---\n\n'.join(body)

#     def _parse_mailbox(self, mailbox):
#         """ don't match
#         """
#         for message in mailbox:
#             body = self._get_body(message)
#             yield parse_body(body)

#     def process_mailbox(self, filename):
#         """
#         """
#         mboxfile = open(filename)
#         mbox = UnixMailbox(mboxfile, message_from_file)
#         msgs = list(self._parse_mailbox(mbox))
#         mboxfile.close()
#         cites = {}
#         for index, msg in enumerate(msgs):
#             for index2, msg2 in enumerate(msgs):
#                 if index == index2 :
#                     continue
#                 for cite in msg.cites:
#                     if ignore_cite(cite):
#                         continue
#                     for para in msg2.content:
#                         if cite in para:
#                             cites.setdefault((index,index2), [])
#                             cites[(index, index2)].append( (cite, para) )
#         return msgs, cites

# if __name__ == '__main__':
#     from pprint import pprint
#     msgs, cites = CitesDetector().process_mailbox(sys.argv[1])
#     for index, msg in enumerate(msgs):
#         print '*'*80, index
#         pprint(msg.content)
#         pprint(msg.cites)
#     print
#     keys = cites.keys()
#     keys.sort()
#     for key in keys:
#         print '='*80, key
#         pprint( cites[key] )



"""hooks triggered on email entities creation:

* look for state change instruction (XXX security)
* set email content as a comment on an entity when comments are supported and
  linking information are found

:organization: Logilab
:copyright: 2003-2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"

import re
import email

from logilab.common.umessage import UMessage
from logilab.mtconverter import TransformError

from cubicweb import UnknownEid
from cubicweb.mail import parse_message_id
from cubicweb.predicates import is_instance
from cubicweb.server import hook


def fix_ownership(cnx, eid, email):
    sender = email.senderaddr.email_of
    if sender and sender.e_schema == 'CWUser' and sender.eid != cnx.user.eid:
        # match a user which is not the connection's user, set owned_by / created_by
        cnx.execute('SET X owned_by U WHERE X eid %(x)s, U eid %(u)s',
                    {'x': eid, 'u': sender.eid})
        cnx.execute('SET X created_by U WHERE X eid %(x)s, U eid %(u)s',
                    {'x': eid, 'u': sender.eid})


class ExtractEmailInformation(hook.Operation):
    """generate a comment on the original entity if supported"""

    def precommit_event(self):
        email = self.email
        # should create a Comment ?
        info = self.info
        origeid = int(info['eid'])
        try:
            origetype = self.cnx.entity_metas(origeid)['type']
        except UnknownEid:
            self.error('email %s is referencing an unknown eid %s',
                       email.messageid, origeid)
            return
        if origetype in self.cnx.vreg.schema['comments'].objects('Comment'):
            try:
                part = email.parts_in_order(prefered_mime_type='text/plain')[0]
            except IndexError:
                pass
            else:
                try:
                    self.insert_comment(origeid, part)
                except Exception:
                    self.exception('while generating comment on %s from email %s',
                                   origeid, email)

    def insert_comment(self, eid, emailpart):
        com = self.cnx.execute(
            'INSERT Comment C: C content %(content)s, '
            'C content_format %(format)s, C comments X, C generated_by E '
            'WHERE X eid %(x)s, E eid %(e)s',
            {'x': eid, 'e': self.email.eid, 'format': u'text/plain',
             'content': emailpart.actual_content()})
        fix_ownership(self.cnx, com[0][0], self.email)


class AnalyzeEmailText(hook.Operation):
    """check if there are some change state instruction in the mail content"""

    def precommit_event(self):
        text = self.email.subject or u''
        for part in self.email.parts_in_order(prefered_mime_type='text/plain'):
            try:
                text += ' ' + part.actual_content()
            except TransformError:
                continue
        # XXX use user session if gpg signature validated
        parser = self.cnx.vreg['components'].select('textanalyzer', self.cnx)
        parser.parse(self, text)

    def fire_event(self, event, evargs):
        if event == 'state-changed':
            evargs['trinfo'].cw_set(generated_by=self.email)
            fix_ownership(self.cnx, evargs['trinfo'].eid, self.email)


class AddEmailCommentHook(hook.Hook):
    """an email has been added, check if associated content should be created
    """
    __regid__ = 'extractmailcontent'
    __select__ = hook.Hook.__select__ & is_instance('Email')
    events = ('after_add_entity',)

    def __call__(self):
        if 'comments' in self._cw.repo.schema:
            for msgid in self.entity.references():
                info = parse_message_id(msgid, self._cw.vreg.config.appid)
                self.info('extracted information from message id %s: %s',
                          msgid, info)
                if info:
                    ExtractEmailInformation(self._cw, email=self.entity, info=info)
                    break
        AnalyzeEmailText(self._cw, email=self.entity)


CLEANUP_RGX = re.compile(r'\bre\s*:', re.I | re.U)


def cleanup_subject(string):
    return CLEANUP_RGX.sub('', string).strip()


class AddEmailPreHook(hook.Hook):
    __regid__ = 'extractmailmetadata'
    __select__ = hook.Hook.__select__ & is_instance('Email')
    events = ('before_add_entity',)

    def address_eid(self, address, alias=None):
        rql = 'Any X WHERE X is EmailAddress, X address LIKE %(addr)s'
        rset = self._cw.execute(rql, {'addr': address})
        if not rset:
            # create a new email address to link to
            alias = alias or None
            # XXX could try to link created address to a person
            eaddress = self._cw.create_entity('EmailAddress', address=address,
                                              alias=alias)
            return eaddress.eid
        # check for a prefered form if any
        return rset.get_entity(0, 0).prefered.eid

    def thread_eid(self, subject, replyeid):
        if replyeid:
            rset = self._cw.execute('EmailThread X WHERE Y in_thread X, Y eid %(y)s',
                                    {'y': replyeid})
            if rset:
                return rset[0][0]
        subject = cleanup_subject(subject)
        thread = self._cw.create_entity('EmailThread', title=subject)
        return thread.eid

    def __call__(self):
        msg = self.entity
        if 'headers' not in msg.cw_edited:
            return
        try:
            message = UMessage(email.message_from_string(msg.headers))
        except Exception:
            self.exception('bad message headers')
            if self._cw.repo.config.mode == 'test':
                raise
            return
        # XXX why limit to a single sender?
        if 'messageid' not in msg.cw_edited:
            msg.cw_edited['messageid'] = message.get('message-id')
        if 'subject' not in msg.cw_edited:
            msg.cw_edited['subject'] = message.get('subject') or u'(no subject)'
        if 'date' not in msg.cw_edited:
            msg.cw_edited['date'] = message.date()
        if 'sender' not in msg.cw_edited:
            try:
                sender = message.multi_addrs('from')[0]
            except IndexError:
                pass
            else:
                sendereid = self.address_eid(sender[1], sender[0])
                msg.cw_edited['sender'] = sendereid
        replyto = message.get('in-reply-to')
        if replyto and 'reply_to' not in msg.cw_edited:
            rset = self._cw.find('Email', messageid=replyto)
            if rset:
                # XXX reply_to should allow multiple objects
                msg.cw_edited['reply_to'] = rset[0][0]
        if 'in_thread' not in msg.cw_edited:
            msg.cw_edited['in_thread'] = self.thread_eid(msg.subject, msg.reply_to and msg.reply_to[0])


class AddEmailPostHook(AddEmailPreHook):
    events = ('after_add_entity',)

    def __call__(self):
        msg = self.entity
        if 'headers' not in msg.cw_edited:
            return
        try:
            message = UMessage(email.message_from_string(msg.headers))
        except Exception:
            return
        if not msg.cc:
            msg.cw_set(cc=[self.address_eid(addr, name)
                           for name, addr in message.multi_addrs('cc')])
        if not msg.recipients:
            msg.cw_set(recipients=set(self.address_eid(addr, name)
                                      for name, addr in message.multi_addrs('to')))
